'''
Created on Feb 15, 2017

@author: julien
'''
import unittest

from minos.experiment.experiment import Experiment, ExperimentParameters,\
    check_experiment_parameters
from minos.experiment.training import Training
from minos.model.design import create_random_blueprint, mix_blueprints,\
    mutate_blueprint
from minos.model.model import Optimizer, Layout
from minos.model.parameter import is_valid_param_value, str_param_name,\
    string_param


class DesignTest(unittest.TestCase):

    def test_random_blueprint(self):
        layout = Layout(
            input_size=100,
            output_size=10,
            output_activation='softmax')
        training = Training(
            objective=None,
            optimizer=None,
            metric=None,
            stopping=None,
            batch_size=None)
        experiment = Experiment(
            'test',
            layout,
            training,
            batch_iterator=None,
            test_batch_iterator=None,
            environment=None,
            parameters=ExperimentParameters(use_default_values=False))
        experiment.parameters.all_search_parameters(True)
        check_experiment_parameters(experiment)
        for _ in range(10):
            blueprint = create_random_blueprint(experiment)
            self.assertIsNotNone(blueprint, 'Should have created a blueprint')

            optimizer = blueprint.training.optimizer
            self.assertIsNotNone(
                optimizer.optimizer,
                'Should have created an optimizer')
            ref_parameters = experiment.parameters.get_optimizers_parameters()
            for name, param in ref_parameters[optimizer.optimizer].items():
                self.assertTrue(
                    is_valid_param_value(param, optimizer.parameters[name]),
                    'Invalid param value')

            self.assertIsNotNone(blueprint.layout, 'Should have created a layout')
            rows = len(blueprint.layout.get_rows())
            self.assertTrue(
                is_valid_param_value(
                    experiment.parameters.get_layout_parameter('rows'),
                    rows),
                'Invalid value')
            for row in blueprint.layout.get_rows():
                blocks = len(row.get_blocks())
                self.assertTrue(
                    is_valid_param_value(
                        experiment.parameters.get_layout_parameter('blocks'),
                        blocks),
                    'Invalid value')
                for block in row.get_blocks():
                    layers = len([
                        l
                        for l in block.get_layers()
                        if l.layer_type != 'Merge'])
                    self.assertTrue(
                        is_valid_param_value(
                            experiment.parameters.get_layout_parameter('layers'),
                            layers),
                        'Invalid value')

    def test_predefined_parameters(self):
        layout = Layout(
            input_size=100,
            output_size=10,
            output_activation='softmax')
        training = Training(
            objective=None,
            optimizer=Optimizer('SGD', {'lr': 1}),
            metric=None,
            stopping=None,
            batch_size=None)
        experiment = Experiment(
            'test',
            layout,
            training,
            batch_iterator=None,
            test_batch_iterator=None,
            environment=None)
        for _ in range(10):
            blueprint = create_random_blueprint(experiment)
            self.assertIsNotNone(blueprint, 'Should have created a blueprint')

            self.assertEqual(
                training.optimizer.optimizer,
                blueprint.training.optimizer.optimizer,
                'Should have created an optimizer')
            self.assertEqual(
                blueprint.training.optimizer.parameters['lr'],
                training.optimizer.parameters['lr'],
                'Should have copied predefined parameter')

    def test_predefined_layer_type(self):
        layout = Layout(
            input_size=100,
            output_size=10,
            output_activation='softmax')
        training = Training(
            objective=None,
            optimizer=Optimizer('SGD', {'lr': 1}),
            metric=None,
            stopping=None,
            batch_size=None)
        experiment = Experiment(
            'test',
            layout,
            training,
            batch_iterator=None,
            test_batch_iterator=None,
            environment=None)
        layer_types = ['Dropout']
        experiment.parameters.block_layer_types(string_param(layer_types))
        for _ in range(10):
            blueprint = create_random_blueprint(experiment)
            self.assertIsNotNone(blueprint, 'Should have created a blueprint')
            for layer in blueprint.layout.get_layers():
                self.assertTrue(
                    layer.layer_type in layer_types,
                    'Should have used predefined layer types')
            

    def test_predefined_layout(self):
        layout = Layout(
            input_size=100,
            output_size=10,
            output_activation='softmax',
            block=[
                ('Dense', {'activation': 'relu'}),
                'Dropout',
                ('Dense', {'output_dim': 100})])
        training = Training(
            objective=None,
            optimizer=Optimizer(),
            metric=None,
            stopping=None,
            batch_size=None)
        experiment = Experiment(
            'test',
            layout,
            training,
            batch_iterator=None,
            test_batch_iterator=None,
            environment=None,
            parameters=ExperimentParameters(use_default_values=False))
        experiment.parameters.search_parameter('layout', False)
        experiment.parameters.search_parameter('parameters', True)
        experiment.parameters.search_parameter('optimizer', True)
        for _ in range(10):
            blueprint1 = create_random_blueprint(experiment)
            blueprint2 = create_random_blueprint(experiment)
            blueprint3 = mix_blueprints(blueprint1, blueprint2, experiment.parameters)
            blueprint4 = mutate_blueprint(blueprint1, experiment.parameters, mutate_in_place=False)
            for idx, blueprint in enumerate([blueprint1, blueprint2, blueprint3, blueprint4]):
                self.assertIsNotNone(blueprint, 'Should have created a blueprint')
                self.assertIsNotNone(blueprint.layout, 'Should have created a layout')
                self.assertEqual(
                    1,
                    len(blueprint.layout.get_rows()),
                    'Should have 1 row')
                self.assertEqual(
                    1,
                    len(blueprint.layout.get_blocks()),
                    'Should have 1 block')
                self.assertEqual(
                    len(layout.block),
                    len(blueprint.layout.get_layers()),
                    'Should have predefined layers count')
                for i, row in enumerate(blueprint.layout.get_rows()):
                    blocks = len(row.get_blocks())
                    self.assertTrue(
                        is_valid_param_value(
                            experiment.parameters.get_layout_parameter('blocks'),
                            blocks),
                        'Invalid value')
                    for block in row.get_blocks():
                        self.assertEqual(
                            len(layout.block),
                            len(block.layers),
                            'Should have used template')
                        for i in range(len(layout.block)):
                            layer = layout.block[i]
                            layer_type = str_param_name(layer[0] if isinstance(layer, tuple) else layer)
                            params = layer[1] if isinstance(layer, tuple) else dict()
                            self.assertEqual(
                                layer_type,
                                block.layers[i].layer_type,
                                'Should have used the predefined layer type')
                            for name, value in params.items():
                                self.assertEqual(
                                    value,
                                    block.layers[i].parameters[name],
                                    'Should have used the predefined parameter value for blueprint %d' % idx)


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



'''
Created on Feb 14, 2017

@author: julien
'''
from copy import deepcopy
from random import Random, shuffle

from minos.experiment.experiment import Blueprint
from minos.experiment.training import Training
from minos.model.model import Layout, Row, Layer, Block,\
    Optimizer
from minos.model.parameter import random_initial_param_value,\
    random_list_element, mutate_param


rand = Random()


def create_random_blueprint(experiment):
    return Blueprint(
        _random_layout(
            experiment.layout,
            experiment.parameters),
        _random_training(experiment))


def _random_training(experiment):
    training = deepcopy(experiment.training)
    if experiment.parameters.is_optimizer_search():
        training.optimizer = _random_optimizer(
            training.optimizer,
            experiment.parameters)
    return training


def _random_optimizer(optimizer, experiment_parameters):
    ref_parameters = experiment_parameters.get_optimizers_parameters()
    optimizer_id = optimizer.optimizer if optimizer else None
    if not optimizer_id:
        optimizers = list(ref_parameters.keys())
        optimizer_id = optimizers[rand.randint(0, len(optimizers) - 1)]
    if not optimizer_id in ref_parameters:
        raise Exception('Unknown optimizer id: "%s"' % optimizer_id)
    param_space = deepcopy(ref_parameters[optimizer_id])
    if optimizer:
        param_space.update(optimizer.parameters)
    parameters = {
        name: random_initial_param_value(param)
        for name, param in param_space.items()}
    return Optimizer(optimizer_id, parameters)


def _random_layout(layout, experiment_parameters):
    layout = Layout(
        layout.input_size,
        layout.output_size,
        layout.output_activation,
        layout.block)
    if experiment_parameters.is_layout_search():
        rows = random_initial_param_value(experiment_parameters.get_layout_parameter('rows'))
        for row_idx in range(rows):
            layout.rows.append(
                _random_layout_row(
                    layout,
                    row_idx,
                    experiment_parameters))
    else:
        layout.rows = [
            Row(blocks=[
                _instantiate_layout_block(
                    layout,
                    0,
                    experiment_parameters)])]
    if experiment_parameters.is_parameters_search():
        _set_layout_random_parameters(layout, experiment_parameters)
    return layout


def _set_layout_random_parameters(layout, experiment_parameters):
    for layer in layout.get_layers():
        _set_layer_random_parameters(layer, experiment_parameters)


def _set_layer_random_parameters(layer, experiment_parameters):
    param_space = deepcopy(experiment_parameters.get_layer_parameters(layer.layer_type))
    param_space.update(layer.parameters)
    layer.parameters = {
        name: random_initial_param_value(param)
        for name, param in param_space.items()}
    layer.apply_constraints()


def _random_layout_row(layout, row_idx,
                       experiment_parameters, init_layer_parameters=False):
    blocks = random_initial_param_value(param=experiment_parameters.get_layout_parameter('blocks'))
    return Row([
        _random_layout_block(
            layout,
            row_idx,
            experiment_parameters,
            init_layer_parameters=init_layer_parameters)
        for _ in range(blocks)])


def _instantiate_layout_block(layout, row_idx, experiment_parameters, init_layer_parameters=False):
    block = Block()
    _setup_block_input(layout, row_idx, block, experiment_parameters)
    block.layers = _create_template_layers(
        layout.block,
        experiment_parameters,
        init_layer_parameters=init_layer_parameters)
    return block


def _random_layout_block(layout, row_idx,
                         experiment_parameters, init_layer_parameters=False):
    if layout.block:
        return _instantiate_layout_block(
            layout,
            row_idx,
            experiment_parameters,
            init_layer_parameters=init_layer_parameters)
    layers = random_initial_param_value(experiment_parameters.get_layout_parameter('layers'))
    template = []
    for _ in range(layers):
        allowed_layers = get_allowed_new_block_layers(template, experiment_parameters)
        if len(allowed_layers) == 0:
            break
        new_layer = random_list_element(allowed_layers)
        template.append(new_layer)
    block = Block()
    _setup_block_input(layout, row_idx, block, experiment_parameters)
    block.layers += _create_template_layers(
        template,
        experiment_parameters,
        init_layer_parameters)
    return block


def _setup_block_inputs(layout, experiment_parameters):
    for row_idx in range(len(layout.rows)):
        for block in layout.rows[row_idx].blocks:
            _setup_block_input(layout, row_idx, block, experiment_parameters)


def _setup_block_input(layout, row_idx, block, experiment_parameters):
    if row_idx == 0:
        if len(block.input_layers) > 0:
            block.input_layers = list()
        return
    previous_row = layout.get_rows()[row_idx - 1]
    if len(previous_row.get_blocks()) == 1:
        if len(block.input_layers) > 0:
            block.input_layers = list()
        return
    if len(block.input_layers) > 0:
        return
    template = layout.block_input
    if not template:
        template = [('Merge', dict(mode='concat'))]
    block.input_layers = _create_template_layers(template, experiment_parameters)
    for layer in block.input_layers:
        layer.apply_constraints()


def _get_layer_type(layer):
    layer_type = layer.layer_type\
        if isinstance(layer, Layer) else layer[0]
    return layer_type


def _is_first_layer_merge(layers):
    if len(layers) == 0:
        return False
    return _is_merge_layer(layers[0])


def _is_merge_layer(layer):
    layer_type = _get_layer_type(layer)
    return layer_type == 'Merge'


def _is_first_layers_match_template(layers, template):
    for idx in range(len(template)):
        if idx >= len(layers):
            return False
        layer_type = layers[idx].layer_type\
            if isinstance(layers[idx], Layer) else layers[idx][0]
        if template[idx][0] != layer_type:
            return False
    return True


def _create_template_layers(template, experiment_parameters, init_layer_parameters=False):
    layers = list()
    for layer_definition in template:
        layer_type = layer_definition[0]\
            if isinstance(layer_definition, tuple) else layer_definition
        parameter_constraints = deepcopy(layer_definition[1])\
            if isinstance(layer_definition, tuple) else dict()
        layer = Layer(
            layer_type,
            parameter_constraints=parameter_constraints)
        if init_layer_parameters:
            _set_layer_random_parameters(layer, experiment_parameters)
        layers.append(layer)
    return layers


def is_allowed_block_layer(layers, new_layer):
    previous_layer_type = None
    if len(layers) > 0:
        if isinstance(layers[-1], Layer):
            previous_layer_type = layers[-1].layer_type
        elif isinstance(layers[-1], tuple):
            previous_layer_type = layers[-1][0]
        else:
            previous_layer_type = layers[-1]
    if new_layer == 'BatchNormalization':
        return previous_layer_type == 'Dense'
    return new_layer != previous_layer_type


def get_allowed_new_block_layers(layers, parameters):
    return [
        new_layer
        for new_layer in parameters.get_block_layer_types()
        if is_allowed_block_layer(layers, new_layer)]


def mutate_blueprint(blueprint, parameters,
                     p_mutate_layout=0.25, p_mutate_param=0.25,
                     layout_mutation_count=1, layout_mutables=None, mutate_in_place=True):
    if not mutate_in_place:
        blueprint = deepcopy(blueprint)
    if parameters.is_layout_search():
        if rand.random() < p_mutate_layout:
            _mutate_layout(
                blueprint.layout,
                parameters,
                mutables=layout_mutables,
                mutation_count=layout_mutation_count)
    if parameters.is_parameters_search():
        _mutate_layer_parameters(
            blueprint.layout,
            parameters,
            p_mutate_param=p_mutate_param)
    if parameters.is_optimizer_search():
        _mutate_optimizer(
            blueprint.training.optimizer,
            parameters,
            p_mutate_param=p_mutate_param)
    return blueprint


def _mutate_layout(layout, parameters,
                   mutables=None, mutation_count=1):
    mutables = mutables or ['rows', 'blocks', 'layers']
    shuffle(mutables)
    for mutation in mutables[:mutation_count]:
        if mutation == 'rows':
            _mutate_layout_rows(layout, parameters)
        if mutation == 'blocks':
            _mutate_layout_blocks(layout, parameters)
        if mutation == 'layers':
            _mutate_layout_layers(layout, parameters)
    _setup_block_inputs(layout, parameters)


def _mutate_layout_rows(layout, parameters):
    current_rows = len(layout.get_rows())
    new_rows = mutate_param(
        parameters.get_layout_parameter('rows'),
        current_rows)
    if new_rows > current_rows:
        for row_idx in range(current_rows, new_rows):
            layout.rows.append(
                _random_layout_row(
                    layout,
                    row_idx,
                    parameters,
                    init_layer_parameters=True))
    else:
        while len(layout.get_rows()) > new_rows:
            idx = rand.randint(0, len(layout.rows) - 1)
            layout.rows = [
                r
                for i, r in enumerate(layout.rows)
                if i != idx]


def _mutate_layout_blocks(layout, parameters):
    row = random_list_element(layout.rows)
    row_idx = layout.rows.index(row)
    current_blocks = len(row.get_blocks())
    new_blocks = mutate_param(
        parameters.get_layout_parameter('blocks'),
        current_blocks)
    if new_blocks > current_blocks:
        for _ in range(current_blocks, new_blocks):
            row.blocks.append(
                _random_layout_block(
                    layout,
                    row_idx,
                    parameters,
                    init_layer_parameters=True))
    else:
        while len(row.blocks) > new_blocks:
            idx = rand.randint(0, len(row.blocks) - 1)
            row.blocks = [
                b
                for i, b in enumerate(row.blocks)
                if i != idx]


def _mutate_layout_layers(layout, parameters):
    row = random_list_element(layout.rows)
    block = random_list_element(row.blocks)
    current_layers = len(block.get_layers())
    new_layers = mutate_param(
        parameters.get_layout_parameter('layers'),
        current_layers)
    if new_layers > current_layers:
        for _ in range(current_layers, new_layers):
            layers = get_allowed_new_block_layers(block.layers, parameters)
            block.layers += _create_template_layers(
                [random_list_element(layers)],
                parameters,
                init_layer_parameters=True)
    else:
        while len(block.layers) > new_layers:
            idx = rand.randint(0, len(block.layers) - 1)
            block.layers = [
                l
                for i, l in enumerate(block.layers)
                if i != idx]


def _mutate_layer_parameters(layout, parameters, p_mutate_param=0.1):
    for row in layout.rows:
        for block in row.blocks:
            for layer in block.layers:
                _mutate_layer(layer, parameters, p_mutate_param)


def _mutate_layer(layer, parameters, p_mutate_param=0.1):
    param_space = deepcopy(parameters.get_layer_parameters(layer.layer_type))
    for name, value in layer.parameters.items():
        if rand.random() < p_mutate_param:
            layer.parameters[name] = mutate_param(param_space[name], value)
    layer.apply_constraints()


def _mutate_optimizer(optimizer, parameters, p_mutate_param=0.1):
    param_space = deepcopy(parameters.get_optimizer_parameters(optimizer.optimizer))
    for name, value in optimizer.parameters.items():
        if rand.random() < p_mutate_param:
            optimizer.parameters[name] = mutate_param(param_space[name], value)


def mix_blueprints(blueprint1, blueprint2, parameters, p_mutate_param=0.05):
    parents = [blueprint1.layout, blueprint2.layout]
    layout = _mix_layouts(parents, parameters, p_mutate_param)
    parents = [blueprint1.training, blueprint2.training]
    training = _mix_trainings(parents, parameters, p_mutate_param)
    return Blueprint(layout, training)


def _mix_trainings(parents, parameters, p_mutate_param=0.05):
    training = Training(
        objective=parents[0].objective,
        optimizer=None,
        metric=parents[0].metric,
        stopping=parents[0].stopping,
        batch_size=parents[0].batch_size)
    parent_optimizers = [p.optimizer for p in parents]
    if parameters.is_optimizer_search():
        training.optimizer = _mix_optimizers(parent_optimizers)
        _mutate_optimizer(training.optimizer, parameters, p_mutate_param)
    else:
        training.optimizer = random_list_element(parent_optimizers)
    return training


def _mix_optimizers(parents):
    return deepcopy(random_list_element(parents))


def _mix_layouts(parent_layouts, parameters, p_mutate_param=0.05):
    layout = Layout(
        input_size=parent_layouts[0].input_size,
        output_size=parent_layouts[0].output_size,
        output_activation=parent_layouts[0].output_activation,
        block=parent_layouts[0].block,
        block_input=parent_layouts[0].block_input)
    if parameters.is_layout_search():
        rows = random_list_element([len(p.rows) for p in parent_layouts])
        for row_idx in range(rows):
            parent_rows = [p.rows[row_idx] for p in parent_layouts if row_idx < len(p.rows)]
            layout.rows.append(_mix_row(layout, row_idx, parent_rows, parameters))
        _setup_block_inputs(layout, parameters)
        if parameters.is_parameters_search():
            _mutate_layer_parameters(
                layout,
                parameters,
                p_mutate_param=p_mutate_param)
    else:
        layout.rows = random_list_element([
            parent_layouts[0].rows,
            parent_layouts[1].rows])
    return layout


def _mix_row(layout, row_idx, parent_rows, parameters):
    row = Row()
    blocks = random_list_element([len(p.blocks) for p in parent_rows])
    for block_idx in range(blocks):
        parent_blocks = [p.blocks[block_idx] for p in parent_rows if block_idx < len(p.blocks)]
        parent = random_list_element(parent_blocks)
        new_block = deepcopy(parent)
        row.blocks.append(new_block)
    return row



'''
Created on Feb 6, 2017

@author: julien
'''
from minos.experiment.experiment import Experiment, ExperimentParameters
from minos.experiment.ga import run_ga_search_experiment
from minos.experiment.training import Training, AccuracyDecreaseStoppingCondition,\
    EpochStoppingCondition
from minos.model.build import ModelBuilder
from minos.model.model import Objective, Optimizer, Metric, Layout
from minos.model.parameter import int_param, string_param, float_param
from minos.train.utils import CpuEnvironment, cpu_device, Environment

from examples.ga.dataset import get_reuters_dataset
import numpy as np
from build.lib.minos.experiment.experiment import load_experiment_best_blueprint


np.random.seed(1337)
max_words = 1000


def build_layout(input_size, output_size):
    """ Here we define a minimal layout. We don't specify the architecture.
    Layouts will be randomly generated using the min and max numbers of rows, blocks and layers
    """
    return Layout(
        input_size=input_size,
        output_size=output_size,
        output_activation='softmax')


def custom_experiment_parameters():
    """ Here we define the experiment parameters.
    We are using use_default_values=True, which will initialize
    all the parameters with their default values. These parameters are then fixed
    for the duration of the experiment and won't evolve.
    That means that we need to manually specify which parametres we want to test,
    and the possible values, either intervals or lists of values.

    If we want to test all the parameters and possible values, we can
    set use_default_values to False. In that case, random values will be generated
    and tested during the experiment. We can redefine some parameters if we want to
    fix their values.
    Reference parameters and default values are defined in minos.model.parameters
    """
    experiment_parameters = ExperimentParameters(use_default_values=True)
    experiment_parameters.layout_parameter('blocks', int_param(1, 5))
    experiment_parameters.layout_parameter('layers', int_param(1, 5))
    experiment_parameters.layer_parameter('Dense.output_dim', int_param(10, 500))
    experiment_parameters.layer_parameter('Dense.activation', string_param(['relu', 'tanh']))
    experiment_parameters.layer_parameter('Dropout.p', float_param(0.1, 0.9))
    return experiment_parameters


def accuracy_decrease_stopping_condition():
    """ This stopping condition lets the training continue as long as the
    accuracy improves and stops if it doesn't improve for 5 epochs
    """
    return AccuracyDecreaseStoppingCondition(
        min_epoch=2,
        max_epoch=10,
        noprogress_count=5)


def epoch_stopping_condition():
    """ This stopping condition lets the training run for 10 epochs
    """
    return EpochStoppingCondition(epoch=10)


def search_model(experiment_label, steps, batch_size=32):
    """ This is where we put everythin together.
    We get the dataset, build the Training and Experiment objects, and run the experiment.
    The experiments logs are generated in ~/minos/experiment_label
    We use the CpuEnvironment to have the experiment run on the cpu, with 2 parralel processes.
    We could use GpuEnvironment to use GPUs, and specify which GPUs to use, and how many tasks
    per GPU
    """
    batch_iterator, test_batch_iterator, nb_classes = get_reuters_dataset(batch_size, max_words)
    layout = build_layout(max_words, nb_classes)
    training = Training(
        Objective('categorical_crossentropy'),
        Optimizer(optimizer='Adam'),
        Metric('categorical_accuracy'),
        epoch_stopping_condition(),
        batch_size)
    parameters = custom_experiment_parameters()
    experiment = Experiment(
        experiment_label,
        layout,
        training,
        batch_iterator,
        test_batch_iterator,
        CpuEnvironment(n_jobs=2),
        parameters=parameters)
    run_ga_search_experiment(
        experiment,
        population_size=100,
        generations=steps,
        resume=False,
        log_level='DEBUG')


def load_best_model(experiment_label, step):
    """ Here we load the blueprints generated during an experiment
    and create the Keras model from the top scoring blueprint
    """
    blueprint = load_experiment_best_blueprint(
        experiment_label,
        step,
        Environment())
    return ModelBuilder().build(
        blueprint,
        cpu_device(),
        compile_model=False)


def main():
    experiment_label = 'reuters_experiment'
    steps = 100
    search_model(experiment_label, steps)
    load_best_model(experiment_label, steps - 1)

if __name__ == '__main__':
    main()



'''
Created on Feb 18, 2017

@author: julien
'''
from genericpath import isfile
from os.path import join
import tempfile
import unittest

from minos.experiment.experiment import ExperimentParameters, Experiment,\
    _assert_valid_training_parameters
from minos.experiment.training import Training, EpochStoppingCondition,\
    AccuracyDecreaseStoppingCondition, AccuracyDecreaseStoppingConditionWrapper
from minos.model.build import ModelBuilder
from minos.model.design import create_random_blueprint
from minos.model.model import Layout, Objective, Optimizer, Metric
from minos.train.trainer import ModelTrainer
from minos.train.utils import default_device, CpuEnvironment, cpu_device
from minos.utils import disable_sysout
from tests.fixtures import get_reuters_dataset


class TrainTest(unittest.TestCase):

    def test_train(self):
        disable_sysout()
        with tempfile.TemporaryDirectory() as tmp_dir:
            batch_size = 50
            batch_iterator, test_batch_iterator, nb_classes = get_reuters_dataset(batch_size, 1000)
            layout = Layout(
                input_size=1000,
                output_size=nb_classes,
                output_activation='softmax')
            training = Training(
                objective=Objective('categorical_crossentropy'),
                optimizer=Optimizer(optimizer='Adam'),
                metric=Metric('categorical_accuracy'),
                stopping=EpochStoppingCondition(10),
                batch_size=batch_size)
            experiment_parameters = ExperimentParameters(use_default_values=True)
            experiment_parameters.layout_parameter('rows', 1)
            experiment_parameters.layout_parameter('blocks', 1)
            experiment_parameters.layout_parameter('layers', 1)
            experiment = Experiment(
                'test__reuters_experiment',
                layout,
                training,
                batch_iterator,
                test_batch_iterator,
                CpuEnvironment(n_jobs=1, data_dir=tmp_dir),
                parameters=experiment_parameters)

            blueprint = create_random_blueprint(experiment)
            model = ModelBuilder().build(blueprint, default_device())
            result = model.fit_generator(
                generator=batch_iterator,
                samples_per_epoch=batch_iterator.samples_per_epoch,
                nb_epoch=10,
                validation_data=test_batch_iterator,
                nb_val_samples=test_batch_iterator.sample_count)
            self.assertIsNotNone(
                result,
                'should have fit the model')
            score = model.evaluate_generator(
                test_batch_iterator,
                val_samples=test_batch_iterator.sample_count)
            self.assertIsNotNone(
                score,
                'should have evaluated the model')

    def test_model_trainer(self):
        disable_sysout()
        with tempfile.TemporaryDirectory() as tmp_dir:
            batch_size = 50
            batch_iterator, test_batch_iterator, nb_classes = get_reuters_dataset(batch_size, 1000)
            layout = Layout(
                input_size=1000,
                output_size=nb_classes,
                output_activation='softmax')
            training = Training(
                objective=Objective('categorical_crossentropy'),
                optimizer=Optimizer(optimizer='Adam'),
                metric=Metric('categorical_accuracy'),
                stopping=EpochStoppingCondition(10),
                batch_size=batch_size)
            experiment_parameters = ExperimentParameters(use_default_values=True)
            experiment_parameters.layout_parameter('rows', 1)
            experiment_parameters.layout_parameter('blocks', 1)
            experiment_parameters.layout_parameter('layers', 1)
            experiment = Experiment(
                'test__reuters_experiment',
                layout,
                training,
                batch_iterator,
                test_batch_iterator,
                CpuEnvironment(n_jobs=1, data_dir=tmp_dir),
                parameters=experiment_parameters)

            blueprint = create_random_blueprint(experiment)
            trainer = ModelTrainer(batch_iterator, test_batch_iterator)
            model_filename = join(tmp_dir, 'model')
            model, history, _duration = trainer.train(
                blueprint,
                cpu_device(),
                save_best_model=True,
                model_filename=model_filename)
            self.assertIsNotNone(
                model,
                'should have fit the model')
            self.assertTrue(isfile(model_filename), 'Should have saved the model')
            self.assertIsNotNone(history, 'Should have the training history')

    def test_early_stopping_condition_test(self):
        disable_sysout()
        with tempfile.TemporaryDirectory() as tmp_dir:
            batch_size = 50
            min_epoch = 10
            max_epoch = 15
            batch_iterator, test_batch_iterator, nb_classes = get_reuters_dataset(batch_size, 1000)
            layout = Layout(
                input_size=1000,
                output_size=nb_classes,
                output_activation='softmax')
            training = Training(
                objective=Objective('categorical_crossentropy'),
                optimizer=Optimizer(optimizer='Adam'),
                metric=Metric('categorical_accuracy'),
                stopping=AccuracyDecreaseStoppingCondition(
                    metric='categorical_accuracy',
                    noprogress_count=2,
                    min_epoch=min_epoch,
                    max_epoch=max_epoch),
                batch_size=batch_size)
            experiment_parameters = ExperimentParameters(use_default_values=True)
            experiment_parameters.layout_parameter('rows', 1)
            experiment_parameters.layout_parameter('blocks', 1)
            experiment_parameters.layout_parameter('layers', 1)
            experiment = Experiment(
                'test__reuters_experiment',
                layout,
                training,
                batch_iterator,
                test_batch_iterator,
                CpuEnvironment(n_jobs=1, data_dir=tmp_dir),
                parameters=experiment_parameters)
            _assert_valid_training_parameters(experiment)

            blueprint = create_random_blueprint(experiment)
            trainer = ModelTrainer(batch_iterator, test_batch_iterator)
            model, history, _duration = trainer.train(blueprint, cpu_device(), save_best_model=False)
            self.assertTrue(len(history.epoch) >= min_epoch, 'Should have trained for at least min epoch')
            self.assertTrue(len(history.epoch) <= max_epoch, 'Should have trained for max epoch')
            self.assertIsNotNone(
                model,
                'should have fit the model')
            score = model.evaluate_generator(
                test_batch_iterator,
                val_samples=test_batch_iterator.sample_count)
            self.assertIsNotNone(
                score,
                'should have evaluated the model')


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



'''
Created on Feb 6, 2017

@author: julien
'''
from copy import deepcopy
import json
import logging
from os import path, makedirs
from os.path import join
import pickle

from minos.experiment.training import AccuracyDecreaseStoppingCondition
from minos.model.parameter import Parameter, str_param_name, expand_param_path
from minos.model.parameters import reference_parameters, get_custom_layers,\
    get_custom_activations
from minos.train.utils import Environment
from minos.utils import setup_logging


class Experiment(object):

    def __init__(self, label, layout=None, training=None,
                 batch_iterator=None, test_batch_iterator=None,
                 environment=None, parameters=None, resume=False):
        self.label = label
        self.layout = layout
        self.training = training
        self.batch_iterator = batch_iterator
        self.test_batch_iterator = test_batch_iterator
        self.environment = environment or Environment()
        self.parameters = parameters or ExperimentParameters()

    def get_experiment_data_dir(self):
        return join(
            self.environment.data_dir,
            self.label)

    def get_log_filename(self):
        return path.join(
            self.get_experiment_data_dir(),
            'experiment.log')

    def get_step_log_filename(self, step):
        return path.join(
            self.get_experiment_data_dir(),
            'experiment.step.%d.log' % step)

    def get_step_data_filename(self, step):
        return path.join(
            self.get_experiment_data_dir(),
            'experiment.step.%d.data' % step)

    def get_checkpoint_filename(self):
        return path.join(
            self.get_experiment_data_dir(),
            'experiment.chkpt')

    def evaluate(self, blueprints):
        from minos.train.trainer import MultiProcessModelTrainer
        model_trainer = MultiProcessModelTrainer(
            self.batch_iterator,
            self.test_batch_iterator,
            self.environment)
        return [
            [result[1]]
            for result in model_trainer.build_and_train_models(blueprints)]


def setup_experiment(experiment, resume=False, log_level='INFO'):
    if not path.exists(experiment.get_experiment_data_dir()):
        makedirs(experiment.get_experiment_data_dir())
    setup_logging(
        experiment.get_log_filename(),
        log_level,
        resume=resume)


def experiment_step_logger(experiment, step, individuals):
    blueprints = [
        individual.to_blueprint()
        for individual in individuals]
    with open(experiment.get_step_log_filename(step), 'w') as generation_file:
        json.dump(
            [blueprint.todict() for blueprint in blueprints],
            generation_file,
            indent=1,
            sort_keys=True)
    with open(experiment.get_step_data_filename(step), 'wb') as generation_file:
        pickle.dump(blueprints, generation_file, -1)

    with open(experiment.get_checkpoint_filename(), 'wb') as checkpoint:
        data = {
            'step': step,
            'blueprints': blueprints}
        pickle.dump(data, checkpoint, -1)


def load_experiment_checkpoint(experiment):
    with open(experiment.get_checkpoint_filename(), 'rb') as checkpoint:
        data = pickle.load(checkpoint)
    return data['step'], data['blueprints']


def run_experiment(experiment, runner,
                   resume=False, log_level='INFO', **params):
    check_experiment_parameters(experiment)
    setup_experiment(experiment, resume, log_level)
    population = None
    population_age = 0
    if resume:
        logging.info('Trying to resume experiment')
        step, population = load_experiment_checkpoint(experiment)
        population_age = step + 1
    runner(
        experiment,
        step_logger=experiment_step_logger,
        population_age=population_age,
        population=population,
        **params)


class InvalidParametersException(Exception):

    def __init__(self, detail):
        super().__init__('Invalid parameters: %s' % detail)


def check_experiment_parameters(experiment):
    _assert_search_parameters_defined(experiment.parameters)
    if not experiment.parameters.is_layout_search()\
            and not experiment.layout.block:
        raise InvalidParametersException(
            'You have to specify a block template '
            + ' if you disable layout search')
    if experiment.parameters.is_layout_search()\
            and not experiment.parameters.is_parameters_search():
        raise InvalidParametersException(
            'If you do a layout search, '
            + ' you have to enable parameters search too')

    _assert_valid_training_parameters(experiment)


def _assert_valid_training_parameters(experiment):
    if not isinstance(
            experiment.training.stopping,
            AccuracyDecreaseStoppingCondition):
        return
    training = experiment.training
    if training.metric.metric != training.stopping.metric:
        raise InvalidParametersException(
            'The same metric must be used for training and early stopping, '
            + '%s != %s' % (
                training.metric.metric,
                training.stopping.metric))


def _assert_search_parameters_defined(experiment_parameters):
    for param_name in reference_parameters['search'].keys():
        value = experiment_parameters.get_search_parameter(param_name)
        if value is None or not isinstance(value, bool):
            raise InvalidParametersException('undefined search parameter: %s' % param_name)


def load_experiment_best_blueprint(experiment_label, environment=Environment()):
    experiment = Experiment(experiment_label, environment=environment)
    last_step, _ = load_experiment_checkpoint(experiment)
    blueprints = list()
    for step in range(last_step):
        blueprint = load_experiment_step_best_blueprint(
            experiment_label,
            step,
            environment=environment)
        if blueprint:
            blueprints.append(blueprint)
    if len(blueprints) == 0:
        return None
    return list(sorted(blueprints, key=lambda b: -b.score[0]))[0]


def load_experiment_step_best_blueprint(experiment_label, step, environment=Environment()):
    blueprints = load_experiment_blueprints(
        experiment_label,
        step,
        environment)
    if len(blueprints) == 0:
        return None
    return list(sorted(blueprints, key=lambda b: -b.score[0]))[0]


def load_experiment_blueprints(experiment_label, step, environment=Environment()):
    experiment = Experiment(experiment_label, environment=environment)
    data_filename = experiment.get_step_data_filename(step)
    with open(data_filename, 'rb') as data_file:
        return pickle.load(data_file)


class ExperimentParameters(object):

    def __init__(self, use_default_values=True):
        self.parameters = deepcopy(reference_parameters)
        self._load_custom_definitions()
        if use_default_values:
            self.parameters = self._init_default_values(self.parameters)

    def _init_default_values(self, node):
        if isinstance(node, Parameter):
            if node.default is not None or node.optional:
                return node.default
        elif isinstance(node, dict):
            for name, value in node.items():
                node[name] = self._init_default_values(value)
        return node

    def _load_custom_definitions(self):
        for name, definition in get_custom_layers().items():
            self.layer_parameter(name, definition[1])
            self.add_block_layer_type(name)

        for name in get_custom_activations().keys():
            self.add_activation(name)

    def get_parameter(self, *path):
        node = self.parameters
        path = expand_param_path(path)
        for elem in path:
            if not elem in node:
                return None
            node = node[elem]
        return node

    def _set_parameter(self, _id, value):
        self.parameters = self._set_node_parameter(
            self.parameters,
            _id,
            value)

    def _set_node_parameter(self, node, path, value):
        node = node or dict()
        path = expand_param_path(path)
        if len(path) > 1:
            node[path[0]] = self._set_node_parameter(
                node[path[0]],
                path[1:],
                value)
        else:
            node[path[0]] = value
        return node

    def is_layout_search(self):
        return self.get_search_parameter('layout')

    def is_parameters_search(self):
        return self.get_search_parameter('parameters')

    def is_optimizer_search(self):
        return self.get_search_parameter('optimizer')

    def get_search_parameter(self, name):
        return self.get_parameter('search', name)

    def search_parameter(self, name, value):
        return self._set_parameter(
            ['search', name],
            value)

    def all_search_parameters(self, value):
        for param_name in self.get_parameter('search').keys():
            self.search_parameter(param_name, value)

    def layout_parameter(self, name, value):
        return self._set_parameter(
            ['layout', name],
            value)

    def get_layout_parameter(self, name):
        return self.get_parameter('layout', name)

    def layer_parameter(self, name, value):
        return self._set_parameter(
            ['layers', name],
            value)

    def add_activation(self, name):
        for _layer_name, parameters in self.get_parameter('layers').items():
            for param_name, param_value in parameters.items():
                if 'activation' == param_name and isinstance(param_value, Parameter):
                    param_value.values.append(name)

    def get_block_layer_types(self):
        return self.get_layout_parameter('block.layer_type').values

    def block_layer_types(self, layer_types):
        self.layout_parameter('block.layer_type', layer_types)

    def add_block_layer_type(self, layer_type):
        layer_type_param = self.get_layout_parameter('block.layer_type')
        layer_type_param.values = list(set(layer_type_param.values + [layer_type]))

    def get_layer_parameter(self, name):
        return self.get_parameter('layers', name)

    def get_layer_parameters(self, layer):
        return self.get_parameter('layers', str_param_name(layer))

    def optimizer_parameter(self, name, value):
        return self._set_parameter(
            ['optimizers', name],
            value)

    def get_optimizer_parameters(self, name):
        return self.get_parameter('optimizers', name)

    def get_optimizers_parameters(self):
        return self.get_parameter('optimizers')


class Blueprint(object):

    def __init__(self, layout, training, score=None):
        self.layout = layout
        self.training = training
        self.score = score

    def todict(self):
        return {
            'layout': self.layout.todict(),
            'training': self.training.todict(),
            'score': self.score}



'''
Created on Feb 15, 2017

@author: julien
'''
import unittest

from minos.experiment.experiment import Experiment, ExperimentParameters,\
    check_experiment_parameters
from minos.experiment.training import Training
from minos.model.design import create_random_blueprint, mix_blueprints
from minos.model.model import Layout


class MixTest(unittest.TestCase):

    def test_mutate_layout(self):
        layout = Layout(
            input_size=100,
            output_size=10,
            output_activation='softmax')
        training = Training(
            objective=None,
            optimizer=None,
            metric=None,
            stopping=None,
            batch_size=None)
        experiment = Experiment(
            'test',
            layout,
            training,
            batch_iterator=None,
            test_batch_iterator=None,
            environment=None,
            parameters=ExperimentParameters(use_default_values=False))
        experiment.parameters.all_search_parameters(True)
        check_experiment_parameters(experiment)
        for _ in range(10):
            blueprint1 = create_random_blueprint(experiment)
            blueprint2 = create_random_blueprint(experiment)
            mutant = mix_blueprints(
                blueprint1,
                blueprint2,
                parameters=experiment.parameters,
                p_mutate_param=0.1)
            blueprints = [blueprint1, blueprint2]
            parent_rows = [
                len(b.layout.rows) for b in blueprints]
            self.assertTrue(
                len(mutant.layout.rows) in parent_rows,
                'Should have used one of the parents')


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



'''
Created on Feb 6, 2017

@author: julien
'''
from copy import deepcopy
import logging
import traceback

import keras
from keras.engine.topology import Input, Merge
from keras.engine.training import Model
from keras.layers.core import Dense
from keras.regularizers import L1L2Regularizer

from minos.model.parameters import is_custom_activation, get_custom_activation,\
    is_custom_layer, get_custom_layers, get_custom_layer


class ModelBuilder(object):

    def __init__(self):
        pass

    def build(self, blueprint, device, compile_model=True):
        model = _build_model(blueprint, device)
        if compile_model:
            model.compile(
                optimizer=_build_optimizer(blueprint.training),
                loss=blueprint.training.objective.objective,
                metrics=[blueprint.training.metric.metric])
        return model


def _build_model(blueprint, device):
    import tensorflow as tf
    with tf.device(device):
        inputs = Input(shape=(blueprint.layout.input_size,))
        row_input = inputs
        for row in blueprint.layout.rows:
            row_input = _build_row_model(row_input, row)
        final_layer_input = _maybe_merge_inputs(row_input)
        predictions = Dense(
            blueprint.layout.output_size,
            activation=blueprint.layout.output_activation)(final_layer_input)
        return Model(input=inputs, output=predictions)


def _build_row_model(inputs, row):
    return [
        _build_block_model(inputs, block)
        for block in row.blocks]


def _build_block_model(inputs, block):
    if isinstance(inputs, list) and len(inputs) == 1:
        inputs = inputs[0]
    if block.input_layers and len(block.input_layers) > 0:
        for layer in block.input_layers:
            inputs = _build_layer_model(inputs, layer)
    for layer in block.layers:
        inputs = _build_layer_model(inputs, layer)
    return inputs


def _maybe_merge_inputs(inputs):
    if isinstance(inputs, list) and len(inputs) > 1:
        return Merge(mode='concat')(inputs)
    elif isinstance(inputs, list) and len(inputs) == 1:
        return inputs[0]
    else:
        return inputs


def _build_layer_model(inputs, layer):
    try:
        parameters = _build_layer_parameters(layer)
        model = _get_layer_model(layer.layer_type)
        return model(**parameters)(inputs)
    except Exception as ex:
        logging.debug(traceback.format_exc())
        raise ex


def _get_layer_model(layer_type):
    if is_custom_layer(layer_type):
        return get_custom_layer(layer_type)[0]
    modules = [keras.layers, keras.layers.normalization]
    for module in modules:
        model = getattr(module, layer_type)
        if model:
            return model
    return None


def _build_layer_parameters(layer):
    parameters = deepcopy(layer.parameters)
    regularizers = [
        'activity_regularizer',
        'b_regularizer',
        'W_regularizer',
        'gamma_regularizer',
        'beta_regularizer']
    for regularizer in regularizers:
        if regularizer in parameters:
            parameters[regularizer] = _get_regularizer(parameters[regularizer])
    activation = parameters.get('activation', None)
    if activation:
        if is_custom_activation(activation):
            parameters['activation'] = get_custom_activation(activation)
    return parameters


def _get_regularizer(regularizer_parameter):
    if regularizer_parameter is None\
            or len(regularizer_parameter) == 0\
            or all(value is None for _, value in regularizer_parameter.items()):
        return None
    l1 = regularizer_parameter.get('l1', 0.)
    l2 = regularizer_parameter.get('l2', 0.)
    return L1L2Regularizer(l1, l2)


def _build_optimizer(training):
    optimizer = getattr(keras.optimizers, training.optimizer.optimizer)
    return optimizer(**training.optimizer.parameters)



'''
Created on Feb 12, 2017

@author: julien
'''
import logging
from multiprocessing import Queue, Process
import numpy
from threading import Thread
from time import time
import traceback

from keras.callbacks import ModelCheckpoint
from keras.models import load_model

from minos.experiment.training import EpochStoppingCondition,\
    AccuracyDecreaseStoppingCondition, AccuracyDecreaseStoppingConditionWrapper,\
    get_associated_validation_metric
from minos.train.utils import is_gpu_device, get_device_idx, get_logical_device
from minos.utils import disable_sysout


class MultiProcessModelTrainer(object):

    def __init__(self, batch_iterator, test_batch_iterator, environment):
        self.batch_iterator = batch_iterator
        self.test_batch_iterator = test_batch_iterator
        self.environment = environment

    def build_and_train_models(self, blueprints):
        logging.debug('Training %d models' % len(blueprints))
        return self._start_training_workers(blueprints)

    def _start_training_workers(self, blueprints):
        try:
            total_n_jobs = sum(self.environment.n_jobs)
            work_queue = Queue(total_n_jobs)
            result_queue = Queue(total_n_jobs)
            self.processes = [
                Process(
                    target=model_training_worker,
                    args=(
                        self.batch_iterator,
                        self.test_batch_iterator,
                        device_id,
                        device,
                        work_queue,
                        result_queue))
                for device_id, device in enumerate(self.environment.devices)
                for _job in range(self.environment.n_jobs[device_id])]
            self.process_count = 0
            for process in self.processes:
                self.process_count += 1
                process.start()

            def _work_feeder():
                count = len(blueprints)
                for i, blueprint in enumerate(blueprints):
                    work_queue.put((i, count, blueprint))
                for _ in range(sum(self.environment.n_jobs)):
                    work_queue.put(None)
            Thread(target=_work_feeder).start()

            results = []
            while self.process_count > 0:
                result = result_queue.get()
                if result:
                    logging.debug(
                        'Blueprint %d: best score %f, epoch %d/%d',
                        result[0],
                        result[1],
                        result[2],
                        result[3])
                    results.append(result)
                else:
                    self.process_count -= 1
            results = list(
                sorted(
                    results,
                    key=lambda e: e[0]))
            return results
        except Exception as ex:
            logging.error(ex)
            logging.error(traceback.format_exc())


class ModelTrainer(object):

    def __init__(self, batch_iterator, test_batch_iterator, debug=False):
        from minos.model.build import ModelBuilder
        self.model_builder = ModelBuilder()
        self.batch_iterator = batch_iterator
        self.test_batch_iterator = test_batch_iterator
        self.debug = debug

    def train(self, blueprint, device,
              save_best_model=False, model_filename=None):
        try:
            model = self.model_builder.build(
                blueprint,
                get_logical_device(device))

            self._setup_tf(device)
            nb_epoch, callbacks = self._get_stopping_parameters(blueprint)
            if save_best_model:
                callbacks.append(self._get_model_save_callback(
                    model_filename,
                    blueprint.training.metric.metric))
            start = time()
            history = model.fit_generator(
                self.batch_iterator,
                self.batch_iterator.samples_per_epoch,
                nb_epoch,
                callbacks=callbacks,
                validation_data=self.test_batch_iterator,
                nb_val_samples=self.test_batch_iterator.sample_count)
            if save_best_model:
                del model
                model = load_model(model_filename)
            return model, history, (time() - start)
        except Exception as ex:
            if self.debug:
                logging.error(ex)
                logging.error(traceback.format_exc())
        try:
            from keras import backend
            backend.clear_session()
        except:
            if self.debug:
                logging.error(ex)
                logging.error(traceback.format_exc())
        return None, None, 0

    def _get_model_save_callback(self, model_filename, metric):
        checkpoint = ModelCheckpoint(
            model_filename,
            monitor=metric,
            save_best_only=True)
        return checkpoint

    def _get_stopping_parameters(self, blueprint):
        if isinstance(blueprint.training.stopping, EpochStoppingCondition):
            nb_epoch = blueprint.training.stopping.epoch
            stopping_callbacks = []
        if isinstance(blueprint.training.stopping, AccuracyDecreaseStoppingCondition):
            nb_epoch = max(
                1,
                blueprint.training.stopping.min_epoch,
                blueprint.training.stopping.max_epoch)
            stopping_callbacks = [
                AccuracyDecreaseStoppingConditionWrapper(blueprint.training.stopping)]
        return nb_epoch, stopping_callbacks

    def _setup_tf(self, device):
        import tensorflow as tf
        config = tf.ConfigProto()
        if is_gpu_device(device):
            config.allow_soft_placement = True
            config.gpu_options.visible_device_list = str(
                get_device_idx(device))
            config.gpu_options.allow_growth = True
        elif hasattr(config, 'gpu_options'):
            config.gpu_options.allow_growth = True
            config.gpu_options.visible_device_list = ''
        from keras import backend
        backend.set_session(tf.Session(config=config))


def model_training_worker(batch_iterator, test_batch_iterator,
                          device_id, device, work_queue, result_queue):
    disable_sysout()
    model_trainer = ModelTrainer(
        batch_iterator,
        test_batch_iterator)
    work = work_queue.get()
    while work:
        try:
            idx, _total, blueprint = work
            model, history, duration = model_trainer.train(
                blueprint,
                device)
            if model and history:
                epoch_total = len(history.epoch)
                val_metric = get_associated_validation_metric(blueprint.training.metric.metric)
                epoch_best = numpy.argmax(history.history[val_metric])
                score = history.history[val_metric][epoch_best]
            else:
                score, epoch_best, epoch_total = 0, 0, 0
            result_queue.put((idx, score, epoch_best, epoch_total, blueprint, duration, device_id))
            work = work_queue.get()
        except Exception as ex:
            logging.error(ex)
    result_queue.put(None)



'''
Created on Feb 14, 2017

@author: julien
'''
from minos.experiment.experiment import Blueprint, Experiment, run_experiment
from minos.model.design import create_random_blueprint, mutate_blueprint,\
    mix_blueprints
from minos.search.ga.ga_search import search, GaExperiment, GaIndividual
from copy import deepcopy


class GaModelExperiment(Experiment, GaExperiment):

    def __init__(self, experiment):
        vars(self).update(vars(experiment))
        self.individual_type = GaBlueprint

    def evaluate(self, individuals):
        blueprints = [
            Blueprint(individual.layout, individual.training)
            for individual in individuals]
        return super().evaluate(blueprints)

    def random_individual(self):
        return GaBlueprint(individual=create_random_blueprint(self))

    def mutate_individual(self, individual, mutate_in_place=False):
        return GaBlueprint(
            individual=mutate_blueprint(
                individual,
                self.parameters,
                mutate_in_place=mutate_in_place))

    def mix_individuals(self, individual1, individual2):
        return GaBlueprint(
            individual=mix_blueprints(
                individual1,
                individual2,
                self.parameters))


class GaBlueprint(Blueprint, GaIndividual):

    def __init__(self, **kwargs):
        individual = kwargs.get('individual', None)
        if individual:
            vars(self).update(vars(individual))
            score = getattr(individual, 'score', None)
            if score:
                self.fitness.values = score
        else:
            vars(self).update(kwargs)

    def to_blueprint(self):
        blueprint = Blueprint(**{
            k: v
            for k, v in vars(self).items()
            if k != 'fitness'})
        fitness = getattr(self, 'fitness', None)
        if fitness:
            blueprint.score = fitness.values
        return blueprint

    @classmethod
    def copy(cls, other_individual):
        return GaBlueprint(deepcopy(other_individual.blueprint))

    def clone(self):
        return GaBlueprint.copy(self)


def run_ga_search_experiment(experiment, population_size=50,
                             generations=100, resume=False, log_level='INFO'):
    run_experiment(
        GaModelExperiment(experiment),
        search,
        resume=resume,
        log_level=log_level,
        population_size=population_size,
        generations=generations)



'''
Created on Feb 14, 2017

@author: julien
'''

from minos.model.parameter import int_param, string_param, boolean_param,\
    float_param


reference_parameters = {
    'search': {
        'layout': boolean_param(default=True),
        'parameters': boolean_param(default=True),
        'optimizer': boolean_param(default=True)},
    'layout': {
        'rows': int_param(lo=1, hi=3, default=1),
        'blocks': int_param(lo=1, hi=5, default=1),
        'layers': int_param(lo=1, hi=5, default=1),
        'block': {
            'input_type': string_param([
                'concat',
                'random+concat',
                'concat+random'], default='concat'),
            'input_size': float_param(default=1.),
            'layer_type': string_param(['Dense', 'Dropout', 'BatchNormalization'])}
    },
    'layers': {
        'Dense': {
            'output_dim': int_param(1, 1000, default=100),
            'init': string_param(
                ['uniform',
                 'lecun_uniform',
                 'normal',
                 # 'identity',
                 # 'orthogonal',
                 'zero',
                 'one',
                 'glorot_normal',
                 'glorot_uniform',
                 'he_normal',
                 'he_uniform'],
                default='glorot_uniform'),
            'activation': string_param(
                ['softmax',
                 'softplus',
                 'softsign',
                 'relu',
                 'tanh',
                 'sigmoid',
                 'hard_sigmoid',
                 'linear'],
                default='relu'),
            'bias': boolean_param(default=True),
            'W_regularizer': {
                'l1': float_param(optional=True, default=None),
                'l2': float_param(optional=True, default=None)},
            'b_regularizer': {
                'l1': float_param(optional=True, default=None),
                'l2': float_param(optional=True, default=None)},
            'activity_regularizer': {
                'l1': float_param(optional=True, default=None),
                'l2': float_param(optional=True, default=None)},
            'W_constraint': string_param(
                ['maxnorm', 'nonneg', 'unitnorm'],
                default=None,
                optional=True),
            'b_constraint': string_param(
                ['maxnorm', 'nonneg', 'unitnorm'],
                default=None,
                optional=True)
        },
        'Dropout': {
            'p': float_param(default=0.75)},
        'Merge': {
            'mode': string_param(
                ['sum', 'mul', 'concat', 'ave', 'cos', 'dot', 'max'],
                default='concat',
                mutable=False)},
        'BatchNormalization': {
            'epsilon': float_param(default=0.001),
            'momentum': float_param(default=0.99),
            'beta_init': string_param(
                ['uniform',
                 'lecun_uniform',
                 'normal',
                 # 'identity',
                 # 'orthogonal',
                 'zero',
                 'one',
                 'glorot_normal',
                 'glorot_uniform',
                 'he_normal',
                 'he_uniform'],
                default='glorot_normal'),
            'gamma_init': string_param(
                ['uniform',
                 'lecun_uniform',
                 'normal',
                 # 'identity',
                 # 'orthogonal',
                 'zero',
                 'one',
                 'glorot_normal',
                 'glorot_uniform',
                 'he_normal',
                 'he_uniform'],
                default='glorot_normal'),
            'gamma_regularizer': {
                'l1': float_param(optional=True, default=None),
                'l2': float_param(optional=True, default=None)},
            'beta_regularizer': {
                'l1': float_param(optional=True, default=None),
                'l2': float_param(optional=True, default=None)}
        }
    },
    'optimizers': {
        'SGD': {
            'lr': float_param(default=1e-3),
            'momentum': float_param(default=0.0),
            'decay': float_param(default=0.0),
            'nesterov': boolean_param(default=False)},
        'RMSprop': {
            'lr': float_param(default=1e-3),
            'rho': float_param(default=0.9),
            'epsilon': float_param(default=1e-08),
            'decay': float_param(default=0.0)},
        'Adagrad': {
            'lr': float_param(default=1e-3),
            'epsilon': float_param(default=1e-08),
            'decay': float_param(default=0.0)},
        'Adadelta': {
            'lr': float_param(default=1e-3),
            'rho': float_param(default=0.9),
            'epsilon': float_param(default=1e-08),
            'decay': float_param(default=0.0)},
        'Adam': {
            'lr': float_param(default=1e-3),
            'beta_1': float_param(default=0.9),
            'beta_2': float_param(default=0.999),
            'epsilon': float_param(default=1e-08),
            'decay': float_param(default=0.0)},
        'Adamax': {
            'lr': float_param(default=1e-3),
            'beta_1': float_param(default=0.9),
            'beta_2': float_param(default=0.999),
            'epsilon': float_param(default=1e-08),
            'decay': float_param(default=0.0)},
        'Nadam': {
            'lr': float_param(default=1e-3),
            'beta_1': float_param(default=0.9),
            'beta_2': float_param(default=0.999),
            'epsilon': float_param(default=1e-08),
            'schedule_decay': float_param(default=4e-3)}},
    'metric': string_param([
        'binary_accuracy',
        'categorical_accuracy',
        'sparse_categorical_accuracy',
        'top_k_categorical_accuracy',
        'mean_squared_error',
        'mean_absolute_error',
        'mean_absolute_percentage_error',
        'mean_squared_logarithmic_error',
        'hinge',
        'squared_hinge',
        'categorical_crossentropy',
        'sparse_categorical_crossentropy',
        'binary_crossentropy',
        'kullback_leibler_divergence',
        'poisson',
        'cosine_proximity',
        'matthews_correlation',
        'precision',
        'fbeta_score',
        'fmeasure'
    ]),
    'objective': string_param([
        'mean_squared_error',
        'mean_absolute_error',
        'mean_absolute_percentage_error',
        'mean_squared_logarithmic_error',
        'squared_hinge',
        'hinge',
        'binary_crossentropy',
        'categorical_crossentropy',
        'sparse_categorical_crossentropy',
        'kullback_leibler_divergence',
        'poisson',
        'cosine_proximity'
    ])
}


custom_layers = dict()
custom_activations = dict()


def is_custom_layer(name):
    return name in custom_layers


def is_custom_activation(name):
    return name in custom_activations


def get_custom_layers():
    return dict(custom_layers)


def get_custom_layer(name):
    return custom_layers.get(name, None)


def get_custom_activations():
    return dict(custom_activations)


def get_custom_activation(name):
    return custom_activations.get(name, None)


def register_custom_layer(name, layer, params=None):
    custom_layers[name] = (layer, params)


def register_custom_activation(name, activation):
    custom_activations[name] = activation



'''
Created on Feb 15, 2017

@author: julien
'''
from os.path import isfile
import tempfile
import unittest

from minos.experiment.experiment import Experiment, ExperimentParameters,\
    load_experiment_blueprints, load_experiment_checkpoint,\
    load_experiment_best_blueprint
from minos.experiment.ga import run_ga_search_experiment
from minos.experiment.training import Training, EpochStoppingCondition
from minos.model.build import ModelBuilder
from minos.model.model import Layout, Objective, Metric, Optimizer
from minos.model.parameter import int_param
from minos.train.utils import CpuEnvironment, cpu_device, Environment
from tests.fixtures import get_reuters_dataset
from minos.utils import disable_sysout


class GaSearchTest(unittest.TestCase):

    def test_ga_search(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            epoch = 3
            generations = 2
            batch_size = 50
            batch_iterator, test_batch_iterator, nb_classes = get_reuters_dataset(batch_size, 1000)
            layout = Layout(
                input_size=1000,
                output_size=nb_classes,
                output_activation='softmax')
            training = Training(
                objective=Objective('categorical_crossentropy'),
                optimizer=Optimizer(optimizer='Adam'),
                metric=Metric('categorical_accuracy'),
                stopping=EpochStoppingCondition(epoch),
                batch_size=batch_size)
            experiment_parameters = ExperimentParameters(use_default_values=False)
            experiment_parameters.layout_parameter('rows', 1)
            experiment_parameters.layout_parameter('blocks', 1)
            experiment_parameters.layout_parameter('layers', 1)
            experiment_parameters.layer_parameter('Dense.output_dim', int_param(10, 500))
            experiment_parameters.all_search_parameters(True)

            experiment_label = 'test__reuters_experiment'
            experiment = Experiment(
                experiment_label,
                layout,
                training,
                batch_iterator,
                test_batch_iterator,
                CpuEnvironment(n_jobs=2, data_dir=tmp_dir),
                parameters=experiment_parameters)
            run_ga_search_experiment(experiment, population_size=2, generations=2)
            self.assertTrue(
                isfile(experiment.get_log_filename()),
                'Should have logged')
            self.assertTrue(
                isfile(experiment.get_step_data_filename(0)),
                'Should have logged')
            self.assertTrue(
                isfile(experiment.get_step_log_filename(0)),
                'Should have logged')
            blueprints = load_experiment_blueprints(
                experiment_label, 0,
                Environment(data_dir=tmp_dir))
            self.assertTrue(
                len(blueprints) > 0,
                'Should have saved/loaded blueprints')
            model = ModelBuilder().build(
                blueprints[0],
                cpu_device())
            disable_sysout()
            model.fit_generator(
                generator=batch_iterator,
                samples_per_epoch=batch_iterator.samples_per_epoch,
                nb_epoch=5,
                validation_data=test_batch_iterator,
                nb_val_samples=test_batch_iterator.sample_count)
            score = model.evaluate_generator(
                test_batch_iterator,
                val_samples=test_batch_iterator.sample_count)
            self.assertTrue(score[1] > 0, 'Should have valid score')

            step, population = load_experiment_checkpoint(experiment)
            self.assertEqual(generations - 1, step, 'Should have loaded checkpoint')
            self.assertIsNotNone(population, 'Should have loaded checkpoint')
            blueprint = load_experiment_best_blueprint(
                experiment.label,
                environment=CpuEnvironment(n_jobs=2, data_dir=tmp_dir))
            model = ModelBuilder().build(
                blueprint,
                cpu_device(),
                compile_model=False)
            self.assertIsNotNone(model, 'Should have loaded and built best model from experiment')

if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



'''
Created on Feb 8, 2017

@author: julien
'''

from inspect import isfunction, isclass
from random import Random

import numpy


rand = Random()


class ParameterConstraint(object):

    levels = ['row', 'block', 'layer']

    def __init__(self, namespace_id, name, value, **kwargs):
        self.namespace_id = namespace_id
        self.name = name
        self.value = value
        for level in self.levels:
            if level in kwargs:
                setattr(self, level, kwargs[level])


class Parameter(object):

    def __init__(self, param_type, values=None,
                 lo=None, hi=None, default=None, optional=False, mutable=True):
        self.param_type = param_type
        self.values = values
        self.lo = lo
        self.hi = hi
        self.default = default
        self.optional = optional
        self.mutable = mutable


class Function(object):

    def __init__(self, func, params):
        self.func = func
        self.params = params


def str_param_name(element):
    if isinstance(element, str):
        return element
    if isclass(element) or isfunction(element):
        return element.__name__
    return str(element)


def param_id(path):
    return '.'.join([str_param_name(e) for e in path])


def param_path(_id):
    return _id.split('.')


def expand_param_path(path):
    path = [
        e
        for p in path
        for e in p.split('.')]
    return path


def boolean_param(default=None, optional=False):
    return Parameter(
        bool,
        (True, False),
        default=default,
        optional=optional)


def float_param(lo=0., hi=1., default=None, values=None, optional=False):
    return Parameter(
        float,
        lo=lo,
        hi=hi,
        values=values,
        default=default,
        optional=optional)


def int_param(lo=0, hi=100, default=1, values=None, optional=False):
    return Parameter(
        int,
        lo=lo,
        hi=hi,
        values=values,
        default=default,
        optional=optional)


def string_param(values, default=None, optional=False, mutable=True):
    return Parameter(
        str,
        values=values,
        default=default,
        optional=optional,
        mutable=mutable)


def param(values, default=None, optional=False):
    return Parameter(
        dict,
        values=values,
        default=default,
        optional=optional)


def func_param(function_definitions, default=None, optional=False):
    return Parameter(
        Function,
        values=function_definitions,
        default=default)


def is_valid_param_value(param, value):
    if not isinstance(value, param.param_type):
        return False
    if param.values is not None and len(param.values) > 0:
        return value in param.values
    if param.lo is not None and value < param.lo:
        return False
    if param.hi is not None and value > param.hi:
        return False
    return True


def mutate_param(param, value):
    try:
        if isinstance(param, dict):
            return {
                name: mutate_param(nested, value.get(name, None))
                for name, nested in param.items()}
        if not isinstance(param, Parameter):
            return value
        if param.optional and value is not None and rand.random() < 0.5:
            return None
        if param.values:
            if len(set(param.values)) == 1:
                return param.values[0]
            element = random_list_element(param.values)
            while element == value:
                element = random_list_element(param.values)
            return element
        elif param.lo == param.hi:
            return value
        value = value or 0
        std = (param.hi - param.lo) / 10
        if param.param_type == int:
            std = max(1, std)
        new_value = value + param.param_type(numpy.random.normal(0, std, 1)[0])
        while not is_valid_param_value(param, new_value) or new_value == value:
            new_value = value + param.param_type(numpy.random.normal(0, std, 1)[0])
        return new_value
    except Exception as ex:
        raise


def _is_optional_param(param):
    return isinstance(param, Parameter) and param.optional


def _random_dict_param_value(param):
    keys = [
        key
        for key in param.keys()
        if not _is_optional_param(param[key])
        or rand.random() < 0.5]
    return {
        key: random_param_value(param[key])
        for key in keys}


def _random_list_param_value(param):
    if len(param) == 0:
        return None
    if len(param) == 1:
        return param[0]
    if len(param) == 2:
        param = Parameter(type(param[0]), lo=param[0], hi=param[1])
    elif len(param) > 2:
        param = Parameter(type(param[0]), values=param)
    return random_param_value(param)


def random_param_value(param):
    if isinstance(param, dict):
        return _random_dict_param_value(param)
    if isinstance(param, list) or isinstance(param, tuple):
        return _random_list_param_value(param)
    if not isinstance(param, Parameter):
        return param
    value = 0
    if param.values:
        value = random_list_element(param.values)
    elif param.lo is not None and param.hi is not None:
        if param.param_type == int:
            value = rand.randint(param.lo, param.hi)
        else:
            value = param.lo + (rand.random() * (param.hi - param.lo))
    elif param.lo is not None:
        value = param.lo * (1 + rand.random())
    elif param.hi is not None:
        value = rand.random() * param.hi
    return param.param_type(value)


def random_initial_param_value(param):
    if isinstance(param, dict):
        return _random_dict_param_value(param)
    if isinstance(param, list) or isinstance(param, tuple):
        return _random_list_param_value(param)
    if not isinstance(param, Parameter):
        return param
    if param.values:
        return random_list_element(param.values)

    initial_value = None
    if param.default is not None:
        initial_value = param.default
    elif param.lo is not None:
        initial_value = param.lo
    else:
        initial_value = 0
    if rand.random() < 0.5:
        return mutate_param(param, initial_value)
    return initial_value


def random_list_element(elements):
    if len(elements) == 0:
        return None
    if len(elements) == 1:
        return elements[0]
    return elements[rand.randint(0, len(elements) - 1)]


def default_param_value(param):
    if isinstance(param, dict):
        return {
            name: default_param_value(value)
            for name, value in param.items()}
    elif isinstance(param, Parameter):
        if param.default is not None or param.optional:
            return param.default
    else:
        return param



'''
Created on Feb 15, 2017

@author: julien
'''
from copy import deepcopy
import unittest

from keras.layers.core import Dense

from minos.experiment.experiment import Experiment, ExperimentParameters,\
    check_experiment_parameters
from minos.experiment.training import Training, EpochStoppingCondition
from minos.model.build import ModelBuilder
from minos.model.design import create_random_blueprint, mix_blueprints,\
    mutate_blueprint
from minos.model.model import Layout, Objective, Metric
from minos.model.parameter import int_param, string_param, float_param
from minos.model.parameters import register_custom_activation,\
    register_custom_layer, reference_parameters
from minos.train.utils import cpu_device


class BuildTest(unittest.TestCase):

    def test_build(self):
        layout = Layout(
            input_size=100,
            output_size=10,
            output_activation='softmax')
        training = Training(
            objective=Objective('categorical_crossentropy'),
            optimizer=None,
            metric=Metric('categorical_accuracy'),
            stopping=EpochStoppingCondition(10),
            batch_size=250)

        experiment_parameters = ExperimentParameters(use_default_values=False)
        experiment_parameters.layout_parameter('blocks', int_param(1, 5))
        experiment_parameters.layout_parameter('layers', int_param(1, 5))
        experiment_parameters.layer_parameter('Dense.output_dim', int_param(10, 500))
        experiment_parameters.layer_parameter('Dense.activation', string_param(['relu', 'tanh']))
        experiment_parameters.layer_parameter('Dropout.p', float_param(0.1, 0.9))
        experiment_parameters.all_search_parameters(True)
        experiment = Experiment(
            'test',
            layout,
            training,
            batch_iterator=None,
            test_batch_iterator=None,
            environment=None,
            parameters=experiment_parameters)
        check_experiment_parameters(experiment)
        for _ in range(5):
            blueprint1 = create_random_blueprint(experiment)
            model = ModelBuilder().build(blueprint1, cpu_device())
            self.assertIsNotNone(model, 'Should have built a model')
            blueprint2 = create_random_blueprint(experiment)
            model = ModelBuilder().build(blueprint2, cpu_device())
            self.assertIsNotNone(model, 'Should have built a model')
            blueprint3 = mix_blueprints(blueprint1, blueprint2, experiment_parameters)
            model = ModelBuilder().build(blueprint3, cpu_device())
            self.assertIsNotNone(model, 'Should have built a model')
            blueprint4 = mutate_blueprint(blueprint1, experiment_parameters, mutate_in_place=False)
            model = ModelBuilder().build(blueprint4, cpu_device())
            self.assertIsNotNone(model, 'Should have built a model')

    def test_build_w_custom_definitions(self):

        def custom_activation(x):
            return x

        register_custom_activation('custom_activation', custom_activation)
        register_custom_layer('Dense2', Dense, deepcopy(reference_parameters['layers']['Dense']))

        layout = Layout(
            input_size=100,
            output_size=10,
            output_activation='softmax',
            block=['Dense2'])
        training = Training(
            objective=Objective('categorical_crossentropy'),
            optimizer=None,
            metric=Metric('categorical_accuracy'),
            stopping=EpochStoppingCondition(5),
            batch_size=250)

        experiment_parameters = ExperimentParameters(use_default_values=False)
        experiment_parameters.layout_parameter('blocks', int_param(1, 5))
        experiment_parameters.layout_parameter('layers', int_param(1, 5))
        experiment_parameters.layer_parameter('Dense2.output_dim', int_param(10, 500))
        experiment_parameters.layer_parameter('Dense2.activation', string_param(['custom_activation']))
        experiment_parameters.layer_parameter('Dropout.p', float_param(0.1, 0.9))
        experiment_parameters.all_search_parameters(True)
        experiment = Experiment(
            'test',
            layout,
            training,
            batch_iterator=None,
            test_batch_iterator=None,
            environment=None,
            parameters=experiment_parameters)
        check_experiment_parameters(experiment)
        for _ in range(5):
            blueprint1 = create_random_blueprint(experiment)
            model = ModelBuilder().build(blueprint1, cpu_device())
            self.assertIsNotNone(model, 'Should have built a model')
            blueprint2 = create_random_blueprint(experiment)
            model = ModelBuilder().build(blueprint2, cpu_device())
            self.assertIsNotNone(model, 'Should have built a model')
            blueprint3 = mix_blueprints(blueprint1, blueprint2, experiment_parameters)
            model = ModelBuilder().build(blueprint3, cpu_device())
            self.assertIsNotNone(model, 'Should have built a model')
            blueprint4 = mutate_blueprint(blueprint1, experiment_parameters, mutate_in_place=False)
            model = ModelBuilder().build(blueprint4, cpu_device())
            self.assertIsNotNone(model, 'Should have built a model')


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



'''
Created on Feb 18, 2017

@author: julien
'''
from multiprocessing import Pool
import tempfile
import unittest

from minos.experiment.experiment import ExperimentParameters, Experiment
from minos.experiment.training import Training, EpochStoppingCondition
from minos.model.design import create_random_blueprint
from minos.model.model import Layout, Objective, Optimizer, Metric
from minos.train.utils import default_device, CpuEnvironment
from minos.utils import disable_sysout
from tests.fixtures import get_reuters_dataset


batch_size = 50
batch_iterator, test_batch_iterator, nb_classes = get_reuters_dataset(batch_size, 1000)


def multiprocess_fit(blueprint):
    disable_sysout()
    from minos.model.build import ModelBuilder
    model = ModelBuilder().build(blueprint, default_device())
    model.fit_generator(
        generator=batch_iterator,
        samples_per_epoch=batch_iterator.samples_per_epoch,
        nb_epoch=10,
        validation_data=test_batch_iterator,
        nb_val_samples=test_batch_iterator.sample_count)


class TrainTest(unittest.TestCase):

    def test_train(self):
        n_jobs = 2
        with Pool(n_jobs) as pool,\
                tempfile.TemporaryDirectory() as tmp_dir:
            layout = Layout(
                input_size=1000,
                output_size=nb_classes,
                output_activation='softmax')
            training = Training(
                objective=Objective('categorical_crossentropy'),
                optimizer=Optimizer(optimizer='Adam'),
                metric=Metric('categorical_accuracy'),
                stopping=EpochStoppingCondition(10),
                batch_size=batch_size)
            experiment_parameters = ExperimentParameters(use_default_values=True)
            experiment_parameters.layout_parameter('rows', 1)
            experiment_parameters.layout_parameter('blocks', 1)
            experiment_parameters.layout_parameter('layers', 1)
            experiment = Experiment(
                'test__reuters_experiment',
                layout,
                training,
                batch_iterator,
                test_batch_iterator,
                CpuEnvironment(n_jobs=1, data_dir=tmp_dir),
                parameters=experiment_parameters)

            blueprints = [
                create_random_blueprint(experiment)
                for _ in range(n_jobs)]
            pool.map(multiprocess_fit, blueprints)


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



'''
Created on Feb 7, 2017

@author: julien
'''
import unittest

from keras.layers.core import Dense

from minos.experiment.experiment import ExperimentParameters, Experiment,\
    check_experiment_parameters, InvalidParametersException
from minos.model.parameter import random_param_value, int_param, float_param,\
    string_param
from minos.model.parameters import reference_parameters,\
    register_custom_activation, register_custom_layer


class ParametersTest(unittest.TestCase):

    def test_parameters(self):
        experiment_parameters = ExperimentParameters(use_default_values=False)
        for layer in reference_parameters['layers'].keys():
            for name, _value in reference_parameters['layers'][layer].items():
                self.assertIsNotNone(
                    experiment_parameters.get_layer_parameter('%s.%s' % (layer, name)),
                    'Parameter %s should exist for layer %s' % (name, layer))

    def test_custom_parameters(self):
        experiment_parameters = ExperimentParameters()
        experiment_parameters.layout_parameter('blocks', int_param(1, 10))
        param = experiment_parameters.get_layout_parameter('blocks')
        self.assertTrue(
            1 == param.lo and 10 == param.hi,
            'Should have set values')
        experiment_parameters.layout_parameter('layers', int_param(1, 3))
        param = experiment_parameters.get_layout_parameter('layers')
        self.assertTrue(
            1 == param.lo and 3 == param.hi,
            'Should have set values')
        experiment_parameters.layer_parameter('Dense.activation', string_param(['relu', 'tanh']))
        param = experiment_parameters.get_layer_parameter('Dense.activation')
        self.assertTrue(
            'relu' == param.values[0] and 'tanh' == param.values[1],
            'Should have set values')

    def test_random_value(self):
        param = int_param(values=list(range(10)))
        val = random_param_value(param)
        self.assertTrue(
            isinstance(val, int),
            'Should be an int')
        self.assertTrue(
            val in param.values,
            'Value should be in predefined values')

        param = float_param(values=[i * 0.1 for i in range(10)])
        val = random_param_value(param)
        self.assertTrue(
            isinstance(val, float),
            'Should be a float')
        self.assertTrue(
            val in param.values,
            'Value should be in predefined values')

        param = float_param(lo=.5, hi=.7)
        for _ in range(100):
            val = random_param_value(param)
            self.assertTrue(
                isinstance(val, float),
                'Should be a float')
            self.assertTrue(
                val <= param.hi and val >= param.lo,
                'Value should be in range')

        param = {
            'a': float_param(optional=False),
            'b': float_param(optional=False)}
        for _ in range(10):
            val = random_param_value(param)
            self.assertTrue(
                isinstance(val, dict),
                'Should be a dict')
            self.assertEqual(
                len(param), len(val),
                'Should respect non optional setting')

        param = {
            'a': float_param(optional=True),
            'b': float_param(optional=True)}
        for _ in range(10):
            val = random_param_value(param)
            self.assertTrue(
                isinstance(val, dict),
                'Should be a dict')
            self.assertTrue(
                len(val) >= 0 and len(val) <= len(param),
                'Should respect non optional setting')

    def test_search_parameters(self):
        experiment = Experiment(
            label='test',
            parameters=ExperimentParameters(use_default_values=False))
        valid_parameters = True
        try:
            check_experiment_parameters(experiment)
        except InvalidParametersException:
            valid_parameters = False
        self.assertFalse(valid_parameters, 'Should have raised an exception')

    def test_custom_definitions(self):

        def custom_activation(x):
            return x

        register_custom_activation('custom_activation', custom_activation)
        register_custom_layer('Dense2', Dense, dict(test='test'))
        experiment_parameters = ExperimentParameters(use_default_values=False)
        custom_params = experiment_parameters.get_layer_parameter('Dense2')
        self.assertIsNotNone(
            custom_params,
            'Should have registered custom layer')
        self.assertTrue(
            'test' in custom_params,
            'Should have registered custom layer params')
        activations = experiment_parameters.get_layer_parameter('Dense.activation')
        self.assertTrue(
            'custom_activation' in activations.values,
            'Should have registered custom_activation')


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



'''
Created on Feb 8, 2017

@author: julien
'''


class Layout(object):

    def __init__(self, input_size, output_size,
                 output_activation, block=None, block_input=None, rows=None):
        self.input_size = int(input_size)
        self.output_size = int(output_size)
        self.output_activation = output_activation
        self.block = block
        self.block_input = block_input
        self.rows = rows or list()

    def get_rows(self):
        return self.rows

    def get_blocks(self):
        return [
            block
            for row in self.rows
            for block in row.get_blocks()]

    def get_layers(self):
        return [
            layer
            for row in self.rows
            for layer in row.get_layers()]

    def todict(self):
        dict_repr = dict(vars(self))
        del dict_repr['block']
        dict_repr['rows'] = [row.todict() for row in self.rows]
        return dict_repr


class Row(object):

    def __init__(self, blocks=None):
        self.blocks = blocks or list()

    def get_blocks(self):
        return self.blocks

    def get_layers(self):
        return [
            layer
            for block in self.blocks
            for layer in block.get_layers()]

    def todict(self):
        return {
            'blocks': [
                block.todict()
                for block in self.blocks]}


class Block(object):

    def __init__(self, layers=None, input_layers=None):
        self.layers = layers or list()
        self.input_layers = input_layers or list()

    def get_layers(self):
        return self.layers

    def todict(self):
        return {
            'input_layers': [
                layer.todict()
                for layer in self.input_layers],
            'layers': [
                layer.todict()
                for layer in self.layers]}


class Layer(object):

    def __init__(self, layer_type,
                 parameters=None, parameter_constraints=None):
        self.layer_type = layer_type
        self.parameters = parameters or dict()
        self.parameter_constraints = parameter_constraints or dict()

    def apply_constraints(self):
        self.parameters.update(self.parameter_constraints)

    def todict(self):
        return dict(vars(self))


class Metric(object):

    def __init__(self, metric):
        self.metric = metric

    def todict(self):
        return dict(vars(self))


class Objective(object):

    def __init__(self, objective):
        self.objective = objective

    def todict(self):
        return dict(vars(self))


class Optimizer(object):

    def __init__(self, optimizer=None, parameters=None):
        self.optimizer = optimizer
        self.parameters = parameters or dict()

    def todict(self):
        return dict(vars(self))



'''
Created on Feb 15, 2017

@author: julien
'''
from copy import deepcopy
import unittest

from keras import activations, backend
from keras.engine.topology import Layer
from keras.layers.core import Dense

from minos.experiment.experiment import Experiment, ExperimentParameters,\
    check_experiment_parameters
from minos.experiment.training import Training, EpochStoppingCondition
from minos.model.build import ModelBuilder
from minos.model.design import create_random_blueprint, mutate_blueprint,\
    mix_blueprints
from minos.model.model import Layout, Objective, Metric
from minos.model.parameter import int_param, string_param, float_param
from minos.model.parameters import register_custom_activation,\
    register_custom_layer, reference_parameters
from minos.train.utils import cpu_device


class MutationTest(unittest.TestCase):

    def test_mutate_layout(self):
        layout = Layout(
            input_size=100,
            output_size=10,
            output_activation='softmax')
        training = Training(
            objective=None,
            optimizer=None,
            metric=None,
            stopping=None,
            batch_size=None)
        experiment = Experiment(
            'test',
            layout,
            training,
            batch_iterator=None,
            test_batch_iterator=None,
            environment=None,
            parameters=ExperimentParameters(use_default_values=False))
        experiment.parameters.all_search_parameters(True)
        check_experiment_parameters(experiment)
        for _ in range(10):
            blueprint = create_random_blueprint(experiment)
            mutant = mutate_blueprint(
                blueprint,
                parameters=experiment.parameters,
                p_mutate_layout=1,
                layout_mutation_count=1,
                layout_mutables=['rows'],
                mutate_in_place=False)
            self.assertTrue(
                len(mutant.layout.rows) != len(blueprint.layout.rows),
                'Should have mutated rows')
            mutant = mutate_blueprint(
                blueprint,
                parameters=experiment.parameters,
                p_mutate_layout=1,
                layout_mutation_count=1,
                layout_mutables=['blocks'],
                mutate_in_place=False)
            self.assertTrue(
                len(mutant.layout.get_blocks()) != len(blueprint.layout.get_blocks()),
                'Should have mutated blocks')
            mutant = mutate_blueprint(
                blueprint,
                parameters=experiment.parameters,
                p_mutate_layout=1,
                layout_mutation_count=1,
                layout_mutables=['layers'],
                mutate_in_place=False)
            self.assertTrue(
                len(mutant.layout.get_layers()) != len(blueprint.layout.get_layers()),
                'Should have mutated layers')

    def test_mutate_parameters(self):
        layout = Layout(
            input_size=100,
            output_size=10,
            output_activation='softmax')
        training = Training(
            objective=None,
            optimizer=None,
            metric=None,
            stopping=None,
            batch_size=None)
        experiment = Experiment(
            'test',
            layout,
            training,
            batch_iterator=None,
            test_batch_iterator=None,
            environment=None,
            parameters=ExperimentParameters(use_default_values=False))
        experiment.parameters.all_search_parameters(True)
        for _ in range(10):
            blueprint = create_random_blueprint(experiment)
            mutant = mutate_blueprint(
                blueprint,
                parameters=experiment.parameters,
                p_mutate_layout=0,
                p_mutate_param=1,
                mutate_in_place=False)

            for row_idx, row in enumerate(mutant.layout.rows):
                for block_idx, block in enumerate(row.blocks):
                    for layer_idx, layer in enumerate(block.layers):
                        original_row = blueprint.layout.rows[row_idx]
                        original_block = original_row.blocks[block_idx]
                        original_layer = original_block.layers[layer_idx]
                        for name, value in layer.parameters.items():
                            if value == original_layer.parameters[name]:
                                pass
                            self.assertTrue(
                                value != original_layer.parameters[name],
                                'Should have mutated parameter')

    def test_mutate_w_custom_definitions(self):

        def custom_activation(x):
            return x

        register_custom_activation('custom_activation', custom_activation)
        register_custom_layer('Dense2', Dense, deepcopy(reference_parameters['layers']['Dense']))

        layout = Layout(
            input_size=100,
            output_size=10,
            output_activation='softmax',
            block=['Dense', 'Dense2'])
        training = Training(
            objective=Objective('categorical_crossentropy'),
            optimizer=None,
            metric=Metric('categorical_accuracy'),
            stopping=EpochStoppingCondition(5),
            batch_size=250)

        experiment_parameters = ExperimentParameters(use_default_values=False)
        experiment_parameters.layout_parameter('blocks', int_param(1, 5))
        experiment_parameters.layout_parameter('layers', int_param(1, 5))
        experiment_parameters.layer_parameter('Dense2.output_dim', int_param(10, 500))
        experiment_parameters.layer_parameter('Dropout.p', float_param(0.1, 0.9))
        experiment_parameters.all_search_parameters(True)
        experiment = Experiment(
            'test',
            layout,
            training,
            batch_iterator=None,
            test_batch_iterator=None,
            environment=None,
            parameters=experiment_parameters)
        check_experiment_parameters(experiment)
        for _ in range(10):
            blueprint = create_random_blueprint(experiment)
            mutant = mutate_blueprint(
                blueprint,
                parameters=experiment.parameters,
                p_mutate_layout=0,
                p_mutate_param=1,
                mutate_in_place=False)
            for row_idx, row in enumerate(mutant.layout.rows):
                for block_idx, block in enumerate(row.blocks):
                    for layer_idx, layer in enumerate(block.layers):
                        original_row = blueprint.layout.rows[row_idx]
                        original_block = original_row.blocks[block_idx]
                        original_layer = original_block.layers[layer_idx]
                        for name, value in layer.parameters.items():
                            self.assertTrue(
                                value != original_layer.parameters[name],
                                'Should have mutated parameter')


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



'''
Created on Feb 12, 2017

@author: julien
'''
from logging.config import dictConfigClass
from os.path import isfile
from posix import remove
import sys


def disable_sysout():
    sys.stdout.write = lambda s: s


def setup_logging(filename, level='INFO', resume=False):
    if not resume and isfile(filename):
        remove(filename)
    print('Logging to %s, level=%s' % (filename, level))
    logging_configuration = {
        'version': 1,
        'formatters': {
            'standard': {
                'format': '%(asctime)s [%(levelname)s] %(name)s: %(message)s',
                'datefmt': '%Y-%m-%d %H:%M:%S'
            },
        },
        'handlers': {
            'file': {
                'level': level,
                'class': 'logging.FileHandler',
                'formatter': 'standard',
                'filename': filename
            }
        },
        'loggers': {
            '': {
                'handlers': ['file'],
                'level': level,
                'propagate': True
            }
        }
    }
    dictConfigClass(logging_configuration).configure()



'''
Created on Feb 8, 2017

@author: julien
'''
from keras.callbacks import EarlyStopping


class Training(object):

    def __init__(self, objective, optimizer, metric, stopping, batch_size):
        self.objective = objective
        self.optimizer = optimizer
        self.metric = metric
        self.stopping = stopping
        self.batch_size = batch_size

    def todict(self):
        return {
            'objective': self.objective.todict(),
            'optimizer': self.optimizer.todict(),
            'metric': self.metric.todict(),
            'stopping': self.stopping.todict(),
            'batch_size': self.batch_size}


class EpochStoppingCondition(object):
    """ Stop training after a fixed number of epochs
        # Arguments
        epoch: number of epochs after which training will be stopped.
    """

    def __init__(self, epoch):
        self.epoch = epoch

    def todict(self):
        return dict(vars(self))


class AccuracyDecreaseStoppingCondition(object):
    """ Stop training when the metric observed has stopped improving.
        # Arguments
        noprogress_count: number of epochs with no improvement
            after which training will be stopped.
        min_epoch: minimum number of epochs.
        max_epoch: maximum number of epochs.
    """

    def __init__(self, max_epoch, metric='accuracy',
                 noprogress_count=3, min_epoch=0):
        self.metric = metric
        self.noprogress_count = noprogress_count
        self.min_epoch = min_epoch
        self.max_epoch = max_epoch
        self.epoch = 0

    def is_min_epoch_defined(self):
        return self.min_epoch and self.min_epoch > 0

    def is_max_epoch_defined(self):
        return self.max_epoch and self.max_epoch > 0

    def is_at_least_min_epoch(self):
        return self.is_min_epoch_defined()\
            and self.epoch >= self.min_epoch

    def is_at_most_max_epoch(self):
        return self.is_max_epoch_defined()\
            and self.epoch <= self.max_epoch

    def todict(self):
        return dict(vars(self))


def get_associated_validation_metric(metric):
    if not metric:
        return None
    if metric.startswith('val_'):
        return metric
    return 'val_%s' % metric


class AccuracyDecreaseStoppingConditionWrapper(EarlyStopping):

    def __init__(self, accuracy_condition):
        super().__init__(
            monitor=get_associated_validation_metric(accuracy_condition.metric),
            patience=accuracy_condition.noprogress_count)
        self.accuracy_condition = accuracy_condition

    def on_epoch_end(self, epoch, logs=None):
        self.accuracy_condition.epoch = epoch
        if self.accuracy_condition.is_min_epoch_defined()\
                and not self.accuracy_condition.is_at_least_min_epoch():
            return
        if self.accuracy_condition.is_max_epoch_defined()\
                and not self.accuracy_condition.is_at_most_max_epoch():
            self.stopped_epoch = epoch
            self.model.stop_training = True
            return
        super().on_epoch_end(epoch, logs)



from setuptools import setup
from setuptools import find_packages


setup(name='pyminos',
      version='0.2.6',
      description='Neural nets hyper parameters & architecture search with genetic algorithms',
      keywords=['keras', 'genetic algorithm', 'neural network', 'deep learning'],
      author='Julien Roch',
      author_email='julien.roch@akalea.com',
      url='https://github.com/guybedo/minos',
      license='Apache',
      setup_requires=[
          'numpy>=1.12'],
      install_requires=[
          'numpy>=1.12',
          'keras',
          'deap>=1.0.2'],
      extras_require={
          'h5py': ['h5py'],
          'tests': ['pytest',
                    'pytest-cov',
                    'pytest-pep8',
                    'pytest-xdist',
                    'python-coveralls',
                    'coverage==3.7.1'],
      },
      packages=find_packages())



'''
Created on Feb 12, 2017

@author: julien
'''
import logging
import numpy
from os import path, makedirs
from os.path import join
from posix import access, W_OK
from random import Random
import traceback

from tensorflow.python.client import device_lib

import tensorflow as tf


random = Random()


def cpu_device():
    return '/cpu:0'


def get_available_gpus():
    try:
        local_device_protos = device_lib.list_local_devices()
        return [x.name for x in local_device_protos if x.device_type == 'GPU']
    except Exception as ex:
        logging.error(
            'Error while trying to list available GPUs: %s' % str(ex))
        return list()


def default_device():
    gpus = get_available_gpus()
    if len(gpus) > 0:
        return gpus[1]
    return cpu_device()


def get_logical_device(physical_device):
    if is_gpu_device(physical_device):
        return '/gpu:0'
    return physical_device


def is_gpu_device(device):
    return device.startswith('/gpu')


def get_device_idx(device):
    return device.split(':')[1]


class Environment(object):

    def __init__(self, devices=None, n_jobs=None,
                 data_dir=None, tf_logging_level=tf.logging.ERROR):
        self.devices = devices
        self.n_jobs = n_jobs
        if devices and n_jobs and not isinstance(n_jobs, list):
            self.n_jobs = [n_jobs for _ in devices]
        self.data_dir = data_dir or self._init_minos_dir()
        self.tf_logging_level = tf_logging_level

    def _init_minos_dir(self):
        base_dir = path.expanduser('~')
        if not access(base_dir, W_OK):
            base_dir = path.join('/tmp')
        minos_dir = join(base_dir, 'minos')
        if not path.exists(minos_dir):
            makedirs(minos_dir)
        return minos_dir


class CpuEnvironment(Environment):

    def __init__(self, n_jobs=1, data_dir=None,
                 tf_logging_level=tf.logging.ERROR):
        super().__init__(
            ['/cpu:0'],
            n_jobs,
            data_dir=data_dir,
            tf_logging_level=tf_logging_level)


class GpuEnvironment(Environment):

    def __init__(self, devices=['/gpu:0'], n_jobs=1, data_dir=None,
                 tf_logging_level=tf.logging.ERROR):
        super().__init__(
            devices,
            n_jobs,
            data_dir=data_dir,
            tf_logging_level=tf_logging_level)


class SimpleBatchIterator(object):

    def __init__(self, X, y, batch_size,
                 X_transform=None, y_transform=None,
                 autoloop=False, autorestart=False, preload=False):
        self.X = X
        self.y = y
        self.batch_size = batch_size
        self.X_transform = X_transform
        self.y_transform = y_transform
        self.autoloop = autoloop
        self.autorestart = autorestart
        self.index = 0
        self.preload = preload
        self.batch_size = batch_size
        self.sample_count = len(X)
        self.samples_per_epoch = self.sample_count
        self.X, self.y = create_batches(
            X,
            y,
            batch_size,
            shuffle=True)

    def _transform_data(self, X, y):
        if self.X_transform:
            X = self.X_transform.fit_transform(X)
        if self.y_transform:
            y = self.y_transform.fit_transform(y)
        return numpy.asarray(X), numpy.asarray(y)

    def __iter__(self):
        return self

    def __next__(self):
        try:
            if self.index >= len(self.X):
                if self.autorestart or self.autoloop:
                    self.index = 0
                if self.autorestart or not self.autoloop:
                    return None
            X, y = self._transform_data(
                self.X[self.index],
                self.y[self.index])
            self.index += 1
            return X, y
        except Exception as ex:
            logging.error('Error while iterating %s' % str(ex))
            try:
                logging.error(traceback.format_exc())
            finally:
                pass
            raise ex


def create_batches(X, y, batch_size, shuffle=True):
    X = [
        X[i:i + batch_size]
        for i in range(0, len(X), batch_size)]
    y = [
        y[i:i + batch_size]
        for i in range(0, len(y), batch_size)]
    if shuffle:
        shuffle_batch(X, y)
    return X, y


def shuffle_batches(X_batches, y_batches):
    for X, y in zip(X_batches, y_batches):
        shuffle_batch(X, y)


def shuffle_batch(X, y):
    for i in range(len(X)):
        swap_idx = random.randint(i, len(X) - 1)
        swap(X, i, swap_idx)
        swap(y, i, swap_idx)


def swap(values, idx1, idx2):
    swap = values[idx2]
    values[idx2] = values[idx1]
    values[idx1] = swap



'''
Created on Dec 5, 2016

@author: julien
'''
import json
import logging
import numpy
from os import path
from random import Random

from deap import creator, base, tools


toolbox = base.Toolbox()
rand = Random()


def random_individual(experiment, individual_type):
    return individual_type(individual=experiment.random_individual())


def make_individual(individual_type, original):
    return individual_type(individual=original)


def mutate(experiment, individual):
    experiment.mutate_individual(individual, mutate_in_place=True)


def mate(experiment, individual1, individual2, count=2):
    children = [
        experiment.mix_individuals(
            toolbox.clone(individual1),
            toolbox.clone(individual2))
        for _ in range(count)]
    return [
        toolbox.make_individual(child)
        for child in children]


def init_ga_env(experiment):
    if not isinstance(experiment, GaExperiment):
        raise Exception(
            'Unexpected experiment type %s'
            % str(experiment))
    creator.create("FitnessMax", base.Fitness, weights=(1.0,))
    creator.create(
        "Individual",
        experiment.individual_type,
        fitness=creator.FitnessMax)  # @UndefinedVariable
    toolbox.register(
        "individual",
        random_individual,
        experiment,
        creator.Individual)  # @UndefinedVariable
    toolbox.register(
        "make_individual",
        make_individual,
        creator.Individual)  # @UndefinedVariable
    toolbox.register("population", tools.initRepeat, list, toolbox.individual)
    toolbox.register("mutate", mutate, experiment)
    toolbox.register("mate", mate, experiment)
    toolbox.register("select", tools.selTournament, tournsize=3)
    toolbox.register("evaluate", experiment.evaluate)


def evolve(population=None, population_size=50,
           population_age=0, generations=50, offspring_count=2, cx_prob=0.5,
           mutpb_prob=0.2, aliens_ratio=0.1, generation_logger=None):
    population = population or toolbox.population(n=population_size)
    for generation in range(population_age, generations):
        logging.info('Evolving generation %d' % generation)
        fit_invalid_individuals(population)
        population = list(sorted(
            population,
            key=lambda i: -i.fitness.values[0]))[:population_size]
        if generation_logger:
            generation_logger(generation, population)
        mates = tools.selBest(population, population_size)
        for ind1, ind2 in zip(mates[::2], mates[1::2]):
            if rand.random() < cx_prob:
                children = toolbox.mate(ind1, ind2, offspring_count)
                for child in children:
                    del child.fitness.values
                population += children
        for mutant in population:
            if rand.random() < mutpb_prob:
                toolbox.mutate(mutant)
                del mutant.fitness.values
        if aliens_ratio > 0:
            population += toolbox.population(n=int(population_size * aliens_ratio))
    return population


def fit_invalid_individuals(population):
    invalid_ind = [ind for ind in population if not ind.fitness.valid]
    if len(invalid_ind) == 0:
        return
    fitnesses = toolbox.evaluate(invalid_ind)
    for ind, fit in zip(invalid_ind, fitnesses):
        if fit is None:
            fit = 0
        ind.fitness.values = fit


def search(experiment, population_size=50, generations=100,
           population_age=0, population=None, log_level='INFO', step_logger=None):
    init_ga_env(experiment)
    if population:
        population = [
            toolbox.make_individual(individual)
            for individual in population]
        if len(population) < population_size:
            population += toolbox.population(n=population_size - len(population))
    evolve(
        population=population,
        population_age=population_age,
        population_size=population_size,
        generations=generations,
        generation_logger=get_generation_logger(experiment, step_logger))


def _get_population_filename(output_dir, experiment_label):
    return path.join(output_dir, '%s.population' % experiment_label)


def get_generation_logger(experiment, step_logger=None):

    def _log(generation, population):
        log_generation_info(generation, population)
        if not step_logger:
            return
        step_logger(experiment, generation, population)

    return _log


def log_generation_info(generation, population):
    sorted_population = list(sorted(
        population,
        key=lambda i: -i.fitness.values[0]))
    population_scores = [
        individual.fitness.values[0]
        for individual in population]
    generation_info = [
        {'generation': generation},
        {'average': numpy.mean(population_scores)},
        {'best_scores': [
            best.fitness.values[0]
            for best in sorted_population[:3]]}]
    logging.info(json.dumps(generation_info))


class GaExperiment(object):

    def __init__(self):
        pass

    def random_individual(self):
        pass


class GaIndividual(object):

    def __init__(self, **kwargs):
        vars(self).update(kwargs)

    def mutate(self):
        pass

    @classmethod
    def mix(cls, individual1, individual2):
        pass

    @classmethod
    def copy(cls, other_individual):
        pass



'''
Created on Feb 19, 2017

@author: julien
'''

from keras.datasets import reuters
from keras.preprocessing.text import Tokenizer
from keras.utils import np_utils
from minos.train.utils import SimpleBatchIterator

import numpy as np


def get_reuters_dataset(batch_size, max_words):
    (X_train, y_train), (X_test, y_test) = reuters.load_data(nb_words=max_words, test_split=0.2)
    nb_classes = np.max(y_train) + 1
    tokenizer = Tokenizer(nb_words=max_words)
    X_train = tokenizer.sequences_to_matrix(X_train, mode='binary')
    X_test = tokenizer.sequences_to_matrix(X_test, mode='binary')
    y_train = np_utils.to_categorical(y_train, nb_classes)
    y_test = np_utils.to_categorical(y_test, nb_classes)

    batch_iterator = SimpleBatchIterator(
        X_train,
        y_train,
        batch_size,
        autoloop=True)
    test_batch_iterator = SimpleBatchIterator(
        X_test,
        y_test,
        len(X_test),
        autoloop=True)
    return batch_iterator, test_batch_iterator, nb_classes